local bars
local Threat = LibStub("Threat-2.0")
local AOE = Omen:NewModule("AOE", Omen.ModuleBase, "AceEvent-3.0", "AceTimer-3.0")
local GUID = LibStub("LibGUIDRegistry-0.1")
local table_sort = _G.table.sort
local math_abs = _G.math.abs
local OL = LibStub("AceLocale-3.0"):GetLocale("Omen")
local L = LibStub("AceLocale-3.0"):GetLocale("Omen_AOE")

local raidTargets, raidTargetsReverse = {}, {}
local playerGUID = UnitGUID("player")

local configOptions = {
	RaidIcons = true,
	ShowTarget = true,
	ShowFocus = true
}
local options = {
	type = "group",
	name = L["AOE Mode"],
	desc = L["AOE Mode"],
	args = {
		test = {
			type = "execute",
			name = OL["Show test bars"],
			desc = OL["Show test bars"],
			func = function() Omen:EnableModule("AOE"); AOE:Test() end
		},
		raidIcons = {
			type = "toggle",
			name = L["Show raid icons"],
			desc = L["Show raid icons"],
			get = function(info) return AOE:GetOption("RaidIcons") end,
			set = function(info, v)
				AOE:SetOption("RaidIcons", v)
				AOE:ReleaseBars()
			end
		},
		showTarget = {
			type = "toggle",
			name = L["Indicate target"],
			desc = L["Indicate target"],
			get = function(info) return AOE:GetOption("ShowTarget") end,
			set = function(info, v)
				AOE:SetOption("ShowTarget", v)
			end
		},
		showFocus = {
			type = "toggle",
			name = L["Indicate focus"],
			desc = L["Indicate focus"],
			get = function(info) return AOE:GetOption("ShowFocus") end,
			set = function(info, v)
				AOE:SetOption("ShowFocus", v)
			end
		}
	}
}

local function sortBars(a, b)
	if a:GetValue() > b:GetValue() then
		return true
	end
	return false
end

function AOE:OnInitialize()
	self:Super("OnInitialize")
	self.icon = select(3, GetSpellInfo(10))
	self:RegisterConfigDefaults(configOptions)
	self:RegisterOptions(options)
end

function AOE:Hint()
	return "AOE Mode\n|cffffffffShows your threat on all targets|r"
end

function AOE:UpdateLayout()
	self:ArrangeBars()
end

function AOE:OnEnable()
	self:Super("OnEnable")
	bars = self.bars
	Threat.RegisterCallback(self, "ThreatUpdated")
	GUID.RegisterCallback(self, "RaidTargetGUIDSet")

	Omen:SetTitle("AOE Mode: " .. UnitName("player"))
	for k, v in Threat:IteratePlayerThreat(playerGUID) do
		AOE:UpdateBar(k, v)
	end	
end

function AOE:OnDisable()
	self:Super("OnDisable")
	GUID.UnregisterCallback(self, "RaidTargetGUIDSet")
	Threat.UnregisterCallback(self, "ThreatUpdated")
end

function AOE.RaidTargetGUIDSet(lib, guid, name, target, unitID)
	if raidTargetsReverse[target] then
		raidTargets[raidTargetsReverse[target]] = nil
	end
	raidTargets[guid] = target
	raidTargetsReverse[target] = guid
end

function AOE:UpdateLayout()
	self:ArrangeBars()
end

local labelsWithIcon = {
	"Icon", "LEFT", 8,
	"Enemy", "LEFT", 52,
	"Threat", "RIGHT", 20,
	"%Max", "RIGHT", 20
}

local labelsNoIcon = {
	"Enemy", "LEFT", 60,
	"Threat", "RIGHT", 20,
	"%Max", "RIGHT", 20
}
			
function AOE:UpdateBar(guid, threat, overrideMax)
	local bar, isNew = self:AcquireBar(guid)
	if isNew then
		bar:SetLabels( unpack(self:GetOption("RaidIcons") and labelsWithIcon or labelsNoIcon) )
		bar:SetLabel("Icon", nil)
		bar:SetLabel("Enemy", GUID:GetName(guid) or "<unknown>")
	end
	local pct = threat / (overrideMax or Threat:GetMaxThreatOnTarget(guid))
	bar:SetLabel("Threat", bar:ConvertToK(threat))
	bar:SetValue(pct)
	self:ArrangeBars()
end

function AOE:ArrangeBars()
	self:ResetBars()
	table_sort(bars, sortBars)
	for i = 1, 80 do		-- A raid of 40 hunters!
		local bar = bars[i]
		if not bar then break end
		bar:SetLabel("Enemy", GUID:GetName(bar.guid) or "<unknown>")
		bar:UnlockHighlight()
		if self:GetOption("ShowFocus") and bar.guid == UnitGUID("focus") then
			bar:SetLabel("Enemy", "[F] " .. (GUID:GetName(bar.guid) or "<unknown>"))
			bar:LockHighlight()
		elseif self:GetOption("ShowTarget") and bar.guid == UnitGUID("target") then
			bar:SetLabel("Enemy", "[T] " .. (GUID:GetName(bar.guid) or "<unknown>"))
			bar:LockHighlight()
		end
		
		if raidTargets[bar.guid] and bar.hasIcon ~= raidTargets[bar.guid] then
			bar:SetLabel("Icon", raidTargets[bar.guid])
		end
		
		bar:LockHighlight()
		if bar:GetValue() > 0 and self:AddBar(bar) then
			local pct = bar:GetValue()
			local r, g, b
			if pct < 0.5 then
				g = 1
				r = (pct / 0.5)
			else
				r = 1
				g = 1 - ((pct - 0.5) / 0.5)
			end
			bar:SetColor(r, g, b)
			bar:SetPercent(pct)
			bar:SetLabel("%Max", ("%2.0f%%"):format(pct*100))
		else
			bar.frame:Hide()
		end
	end
end

function AOE:ThreatUpdated(event, srcGUID, dstGUID, threat)
	if self.testing then
		self:ReleaseBars()
		self.testing = false
	end
	self:UpdateBar(dstGUID, Threat:GetThreat(playerGUID, dstGUID))
end

function AOE:Test()
	self.testing = true
	self:ReleaseBars()

	for i = 1, 25 do
		raidTargets[i] = mod(i, 8) + 1
		self:UpdateBar(i, i, 25)
	end
end
