local bars
local SingleTarget = Omen:NewModule("SingleTarget", Omen.ModuleBase, "AceEvent-3.0", "AceTimer-3.0")
local Threat = LibStub("Threat-2.0")
local GUID = LibStub("LibGUIDRegistry-0.1")
local timers = {}
local table_sort = _G.table.sort
local math_abs = _G.math.abs
local OL = LibStub("AceLocale-3.0"):GetLocale("Omen")
local L = LibStub("AceLocale-3.0"):GetLocale("Omen_SingleTarget")

local configOptions = {
	TPS = {
		UpdateFrequency = 0.1
	},
	Warnings = {
		Enable = true,
		EnableSound = true,
		EnableFlash = true,
		EnableShake = false,
		EnableMessage = true,
		Threshold = 90,
		SoundFile = "Fel Nova"
	},
	AggroBar = {
		Enable = true,
		Color = {r = 1, g = 0, b = 0, a = 1}
	},
	ShowSelf = true
}
local Media = LibStub("LibSharedMedia-3.0")
local sounds = Media:List("sound")

local function GetLSMIndex(t, value)
	for k, v in pairs(Media:List(t)) do
		if v == value then
			return k
		end
	end
	return nil
end

local options = {
	type = "group",
	name = L["Single Target Mode"],
	desc = L["Single Target Mode"],
	args = {
		test = {
			type = "execute",
			name = OL["Show test bars"],
			desc = OL["Show test bars"],
			func = function() Omen:EnableModule("SingleTarget"); SingleTarget:Test() end
		},	
		warnings = {
			type = "group",
			name = L["Warnings"],
			desc = L["Warnings"],
			args = {
				enable = {
					type = "toggle",
					name = L["Enable Warnings"],
					desc = L["Enable Warnings"],
					order = 1,
					get = function(info)
						return SingleTarget:GetOption("Warnings.Enable")
					end,
					set = function(info, v)
						SingleTarget:SetOption("Warnings.Enable", v)
					end
				},
				sound = {
					type = "toggle",
					name = L["Enable Sound"],
					desc = L["Enable Sound"],
					get = function(info)
						return SingleTarget:GetOption("Warnings.EnableSound")
					end,
					set = function(info, v)
						SingleTarget:SetOption("Warnings.EnableSound", v)
					end,
					disabled = function() return not SingleTarget:GetOption("Warnings.Enable") end
				},
				flash = {
					type = "toggle",
					name = L["Enable Screen Flash"],
					desc = L["Enable Screen Flash"],
					get = function(info)
						return SingleTarget:GetOption("Warnings.EnableFlash")
					end,
					set = function(info, v)
						SingleTarget:SetOption("Warnings.EnableFlash", v)
					end,
					disabled = function() return not SingleTarget:GetOption("Warnings.Enable") end
				},
				flash = {
					type = "toggle",
					name = L["Enable Screen Shake"],
					desc = L["Enable Screen Shake"],
					get = function(info)
						return SingleTarget:GetOption("Warnings.EnableShake")
					end,
					set = function(info, v)
						SingleTarget:SetOption("Warnings.EnableShake", v)
					end,
					disabled = function() return not SingleTarget:GetOption("Warnings.Enable") end
				},				
				message = {
					type = "toggle",
					name = L["Enable Warning Message"],
					desc = L["Print a message to screen when you accumulate too much threat"],
					get = function(info)
						return SingleTarget:GetOption("Warnings.EnableMessage")
					end,
					set = function(info, v)
						SingleTarget:SetOption("Warnings.EnableMessage", v)
					end,
					disabled = function() return not SingleTarget:GetOption("Warnings.Enable") end				
				},
				soundFile = {
					type = "select",
					name = L["Sound to play"],
					desc = L["Sound to play"],
					values = sounds,
					get = function(info) return GetLSMIndex("sound", SingleTarget:GetOption("Warnings.SoundFile")) end,
					set = function(info, v)
						SingleTarget:SetOption("Warnings.SoundFile", sounds[v])
						PlaySoundFile(Media:Fetch("sound", sounds[v]))
					end,
					disabled = function() return not SingleTarget:GetOption("Warnings.Enable") end
				},
				threshold = {
					type = "range",
					name = L["Warning Threshold"],
					desc = L["Warning Threshold"],
					get = function(info) return SingleTarget:GetOption("Warnings.Threshold") end,
					set = function(info, v)
						SingleTarget:SetOption("Warnings.Threshold", v)
						SingleTarget.WarningThreshold = v
					end,
					step = 1,
					bigStep = 5,
					min = 60,
					max = 130,
					disabled = function() return not SingleTarget:GetOption("Warnings.Enable") end
				},
				test = {
					type = "execute",
					name = L["Test warnings"],
					desc = L["Test warnings"],
					func = function() SingleTarget:TestWarnings() end,
					order = 101,
					disabled = function() return not SingleTarget:GetOption("Warnings.Enable") end
				}
			}
		},
		showself = {
			type = "toggle",
			name = L["Always show self"],
			desc = L["Always show yourself on the meter"],
			get = function(info) return SingleTarget:GetOption("ShowSelf") end,
			set = function(info, v)
				SingleTarget:SetOption("ShowSelf", v)
				SingleTarget.showSelf = v
			end
		},
		aggrobar = {
			type = "group",
			name = L["Aggro Bars"],
			desc = L["Aggro Bars"],
			args = {
				show = {
					type = "toggle",
					name = L["Show bar"],
					desc = L["Show bar"],
					get = function(info) return SingleTarget:GetOption("AggroBar.Enable") end,
					set = function(info, v)
						SingleTarget:SetOption("AggroBar.Enable", v)
						SingleTarget.showAggroBar = v
						SingleTarget:ArrangeBars()
					end
				},
				color = {
					type = "color",
					name = L["Color"],
					desc = L["The color of the aggro gain bar"],
					get = function(info)
						local c = SingleTarget:GetOption("AggroBar.Color")
						return c.r, c.g, c.b, c.a
					end,
					set = function(info, r, g, b, a)
						SingleTarget:SetOption("AggroBar.Color.r", r)
						SingleTarget:SetOption("AggroBar.Color.g", g)
						SingleTarget:SetOption("AggroBar.Color.b", b)
						SingleTarget:SetOption("AggroBar.Color.a", a)
						local bar = SingleTarget:AcquireBar("AGGRO", true)
						if bar then
							bar:SetColor(r,g,b, a)
						end
					end				
				}
			}
		},
		
	}
}

local function sortBars(a, b)
	if a:GetValue() > b:GetValue() then
		return true
	end
	return false
end
local selfThreat, aggroThreat, aggroGUID = 0, nil, nil
local hostileGUID, hostileUnit, playerGUID

function SingleTarget:OnInitialize()
	self:Super("OnInitialize")
	self.icon = select(3, GetSpellInfo(53))
	self:RegisterConfigDefaults(configOptions)
	self:RegisterOptions(options)
	self.threatUpdatedAt = 0
	self.WarningThreshold = self:GetOption("Warnings.Threshold")
	self.showSelf = self:GetOption("ShowSelf")
	self.showAggroBar = self:GetOption("AggroBar.Enable")
end

function SingleTarget:OnEnable()
	self:Super("OnEnable")
	bars = self.bars
	Threat.RegisterCallback(self, "ThreatUpdated")
	self:RegisterEvent("PLAYER_TARGET_CHANGED")
	self:RegisterEvent("UNIT_TARGET")
	playerGUID = GUID:GetGUID("player")
	
	self:PLAYER_TARGET_CHANGED()
	if timers.OmenCancelTPSUpdate then
		self:CancelTimer(timers.OmenCancelTPSUpdate, true)
	end
	timers.OmenCancelTPSUpdate = nil
end

function SingleTarget:OnDisable()
	self:Super("OnDisable")
	Threat.UnregisterCallback(self, "ThreatUpdated")
end

function SingleTarget:Hint()
	return "Single Target Mode\n|cffffffffShows raid threat on your target|r"
end

function SingleTarget:UpdateLayout()
	self:ArrangeBars()
end

function SingleTarget:UpdateBar(guid, threat)
	local bar, isNew = self:AcquireBar(guid)
	if isNew then
		bar:SetLabels(
			"Name", "LEFT", 40,
			"TPS", "RIGHT", 20,
			"Threat", "RIGHT", 20,
			"%", "RIGHT", 20
		)
		bar:SetClass()
		bar:SetLabel("Name", GUID:GetName(guid) or "<unknown>")
	end
	aggroThreat = aggroThreat or Threat:GetMaxThreatOnTarget(guid)
	bar:SetLabel("Threat", bar:ConvertToK(threat))
	bar:SetValue(threat)
	self:ArrangeBars()
end

function SingleTarget:ArrangeBars()
	local aggroBar
	if #bars > 0 and self.showAggroBar then
		aggroThreat = aggroThreat or 1
		local bar, isNew = self:AcquireBar("AGGRO")
		if isNew then
			bar:SetLabels(
			"Name", "LEFT", 40,
			"TPS", "RIGHT", 20,
			"Threat", "RIGHT", 20,
			"%", "RIGHT", 20
			)
			bar:SetColor(
				self:GetOption("AggroBar.Color.r"),
				self:GetOption("AggroBar.Color.g"),
				self:GetOption("AggroBar.Color.b"),
				self:GetOption("AggroBar.Color.a")
			)
			bar:SetLabel("Name", ("> %s <"):format(L["Pull Aggro"]))
			bar:SetLabel("TPS", "")
		end
		if Threat:UnitInMeleeRange("target") then
			bar:SetValue(aggroThreat * 1.1)
			bar:SetLabel("Threat", bar:ConvertToK(aggroThreat * 1.1))
		else
			bar:SetValue(aggroThreat * 1.3)
			bar:SetLabel("Threat", bar:ConvertToK(aggroThreat * 1.3))
		end
		aggroBar = bar
	else
		local b = self:AcquireBar("AGGRO", true) 
		if b then
			self:ReleaseBar(b)
		end		
	end
	
	local showSelf, selfShown = self.showSelf, false
	local barsShown = 0
	
	self:ResetBars()
	table_sort(bars, sortBars)
	local maxThreat = nil
	for i = 1, 80 do		-- A raid of 40 hunters!
		local bar = bars[i]
		if not bar then break end
		
		if bar.isPet and not Omen:GetOption("Classes.PET") then return end
		if bar.class and not Omen:GetOption("Classes." .. bar.class) then return end
		
		maxThreat = maxThreat or bar:GetValue()
		if bar.guid == playerGUID then
			selfShown = true
		end
		if bar:GetValue() > 0 and (selfShown or self:BarSpaceLeft()) and self:AddBar(bar) then
			local pct = bar:GetValue() / maxThreat
			local visPct = bar:GetValue() / aggroThreat
			bar:SetPercent(pct)
			bar:SetLabel("%", ("%+2.1f%%"):format((visPct-1)*100))
			barsShown = barsShown + 1
		else
			bar.frame:Hide()
		end
	end
	if barsShown == 1 and aggroBar then
		aggroBar.frame:Hide()
	end
end

function SingleTarget:ThreatUpdated(event, srcGUID, dstGUID, threat)
	if not hostileUnit then return end
	if self.testing then
		self:ReleaseBars()
		self.testing = false
	end
	
	if (hostileGUID and dstGUID == hostileGUID) then
		self:CheckWarn(srcGUID, dstGUID, threat)
		self:UpdateBar(srcGUID, threat)
	end
	
	self.threatUpdatedAt = GetTime()
	if not timers.OmenCancelTPSUpdate then
		timers.OmenCancelTPSUpdate = self:ScheduleRepeatingTimer("CancelTPSUpdate", 6)
		timers.OmenTPSUpdate = self:ScheduleRepeatingTimer("UpdateTPS", self:GetOption("TPS.UpdateFrequency"))
	end
end

function SingleTarget:ReInitTPSTimer()
	if timers.OmenTPSUpdate then
		self:CancelTimer(timers.OmenTPSUpdate)
		timers.OmenTPSUpdate = self:ScheduleRepeatingTimer("UpdateTPS", self:GetOption("TPS.UpdateFrequency"))
	end
end

function SingleTarget:CancelTPSUpdate()
	if GetTime() - self.threatUpdatedAt > 5 then
		if timers.OmenTPSUpdate then
			self:CancelTimer(timers.OmenTPSUpdate, true)
			timers.OmenTPSUpdate = nil
		end
		self:CancelTimer(timers.OmenCancelTPSUpdate)
		timers.OmenCancelTPSUpdate = nil
	end
end

function SingleTarget:UpdateTPS()
	local target_guid = GUID:GetGUID(hostileUnit)
	if not target_guid then return end
	for k, bar in pairs(bars) do
		if bar.frame:IsVisible() and bar.guid ~= "AGGRO" then
			bar:SetLabel("TPS", ("%2.0f"):format(Threat:GetTPS(bar.guid, target_guid)))
		end
	end
end

function SingleTarget:DisplayThreatForGUID(guid)
	for k, v in Threat:IterateGroupThreatForTarget(guid) do
		self:UpdateBar(k, v)
	end
end
	
function SingleTarget:PLAYER_TARGET_CHANGED()
	hostileGUID = nil
	hostileUnit = nil
	aggroGUID = nil
	aggroThreat = nil
	if UnitCanAttack("player", "target") then
		hostileUnit = "target"
		self.aggroUnit = "targettarget"
	elseif UnitCanAttack("player", "targettarget") then
		hostileUnit = "targettarget"
		self.aggroUnit = "targettargettarget"
	end
	if not hostileUnit then
		self:ClearBars()
		self:SetTitle(nil)
		return
	end
	hostileGUID = UnitGUID(hostileUnit) 
	
	aggroThreat = Threat:GetThreat(UnitGUID(self.aggroUnit), hostileUnit)
	selfThreat = Threat:GetThreat(UnitGUID("player"), hostileUnit)
	self:DisplayThreatForGUID(hostileGUID)
	self:UpdateTPS()
	self:SetTitle(UnitName(hostileUnit))
	
	self:UNIT_TARGET(nil, "target")
end

function SingleTarget:CheckWarn(s, d, t)
	if s ~= playerGUID and s ~= aggroGUID then return end
	-- if self.playerGUID = self.aggroGUID then return end
	
	if s == playerGUID then
		selfThreat = t
		aggroThreat = aggroGUID and Threat:GetThreat(aggroGUID, d) or Threat:GetMaxThreatOnTarget(d)
	else
		aggroThreat = t
		selfThreat = Threat:GetThreat(playerGUID, d)
	end
	selfThreat = selfThreat or 0
	aggroThreat = max(1, aggroThreat)
	
	local pct = selfThreat / aggroThreat * 100
	if pct > self.WarningThreshold and not self.warned and aggroGUID and playerGUID ~= aggroGUID and aggroThreat > 5 then
		local msg = nil
		if self:GetOption("Warnings.EnableMessage") then
			msg = ("Passed %s%% of %s's threat!"):format(self.WarningThreshold, GUID:GetName(aggroGUID))
		end
		Omen:Warn(
			self:GetOption("Warnings.EnableSound") and Media:Fetch("sound", self:GetOption("Warnings.SoundFile")),
			self:GetOption("Warnings.EnableFlash"),
			self:GetOption("Warnings.EnableShake"),
			msg
		)
		self.warned = true
	elseif pct < (self.WarningThreshold - 5) and self.warned then
		self.warned = false
	end
end

function SingleTarget:TestWarnings()
	Omen:Warn(
		SingleTarget:GetOption("Warnings.EnableSound") and Media:Fetch("sound", SingleTarget:GetOption("Warnings.SoundFile")),
		SingleTarget:GetOption("Warnings.EnableFlash"),
		SingleTarget:GetOption("Warnings.EnableShake"),
		self:GetOption("Warnings.EnableMessage") and "Warning Test" or nil
	)
end

function SingleTarget:SwitchTarget()
	local n = UnitName("targettarget")
	if n then
		local g = GUID:GetGUID(n)
		aggroGUID = g
	end
	timers.TargetSwitch = nil
end

function SingleTarget:UNIT_TARGET(event, unit)
	if not hostileUnit then return end
	if unit == hostileUnit then
		if timers.TargetSwitch then
			self:CancelTimer(timers.TargetSwitch, true)
			timers.TargetSwitch = nil
		end
		if not UnitCastingInfo("target") or not aggroGUID then
			timers.TargetSwitch = self:ScheduleTimer("SwitchTarget", 0.5)
		end
	end
end

function SingleTarget:Test()
	self:ReleaseBars()
	
	aggroThreat = 0 
	self.testing = true
	for i = 1, 25 do
		local threat = random() * 10000
		aggroThreat = max(aggroThreat, threat)
		self:UpdateBar(i, threat)
	end
end
