local MINOR_VERSION = tonumber(("$Revision: 55025 $"):match("%d+"))
local PlayerName = UnitName("player")
local Threat = LibStub("Threat-2.0")
local math_floor = math.floor
local timers = {}
local AceConfig = LibStub("AceConfigDialog-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("Omen")

Omen = LibStub("AceAddon-3.0"):NewAddon("Omen", "AceEvent-3.0", "AceTimer-3.0", "AceBucket-3.0", "AceConsole-3.0", "LibSink-2.0")
Omen:SetDefaultModuleState(false)

function Omen:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("Omen2DB")
	self.db:RegisterDefaults({
		profile = {
			activeSkin = {},
			options = {},
			sinkOptions = {}
		}
	})
	self:SetSinkStorage(self.db.profile.sinkOptions)
	self.configOptions.args.output = self:GetSinkAce3OptionsDataTable()
	self:CreateFrame()
	local active = Omen:GetOption("ActiveWhenSolo")
	if active then
		Threat:RequestActiveOnSolo(active)
	end
	self:Toggle(self:GetOption("Show"))
end

function Omen:OnEnable()
	self:LayoutModuleIcons()
	local module = self:GetOption("Modules.LastActive")
	if module then
		local m = self:GetModule(module, true)
		if m then
			m:Enable()
			return
		end
	end
	-- Enable the first module. Is there a better way?
	for k, v in self:IterateModules() do
		self:EnableModule(k)
		break
	end
end

function Omen:OnDisable()
end

function Omen:Test()
	self = Omen
	if self.activeModule then
		self.activeModule:Test()
	end
end

function Omen:Flash(self, warningLevel)
	UIFrameFlash(LowHealthFrame, 0.15, 0.75, 2.6, false, 0, 0.4);
end

function Omen:Toggle(setting)
	self:SetOption("Show", setting or not Omen.Anchor:IsVisible())
	if self:GetOption("Show") then
		self.Anchor:Show()
	else
		self.Anchor:Hide()
	end
end

--------------------------------
-- Yoinked from BigWigs
--------------------------------

-- Shake properties
local shaking = nil
local SHAKE_DURATION = 0.8
local SHAKE_X = 10
local SHAKE_Y = 10
local fail = nil
local shaker = nil

local originalPoints = nil
local function startShake()
	if not shaking then
		-- store old worldframe positions, we need them all, people have frame modifiers for it etc.
		if not originalPoints then
			originalPoints = {}
			for i = 1, WorldFrame:GetNumPoints() do
				table.insert(originalPoints, {WorldFrame:GetPoint(i)})
			end
		end
		shaking = SHAKE_DURATION -- don't think we want to make this a setting.
		shaker:Show()
	end
end

local function shakeOnUpdate(frame, elapsed)
	shaking = shaking - elapsed
	local x, y = 0, 0 -- Resets to original position if we're supposed to stop.
	if shaking <= 0 then -- stop shaking
		shaking = nil
		shaker:Hide()
	else
		x = math.random(-SHAKE_X,SHAKE_X)
		y = math.random(-SHAKE_Y,SHAKE_Y)
	end
	WorldFrame:ClearAllPoints()
	for i, v in ipairs(originalPoints) do
		WorldFrame:SetPoint(v[1], v[2], v[3], v[4] + x, v[5] + y)
	end
end

--------------------------------
-- End yoink
--------------------------------

function Omen:Shake()
	if not shaker then
		shaker = CreateFrame("Frame", "OmenShaker", UIParent)
		shaker:Hide()
		shaker:SetScript("OnUpdate", shakeOnUpdate)
	end
	local check = WorldFrame:IsProtected()
	if not check then
		startShake()
	elseif check and not fail then
		Omen:Print(("|cffff0000%s:|r %s"):format(L["Error"], L["Unable to use shake warning. If you have nameplates turned on, turn them off to enable shake warnings."]))
		fail = true
	end
end

function Omen:Warn(sound, flash, shake, message)
	if sound then
		PlaySoundFile(sound)
	end
	if flash then self:Flash() end
	if shake then self:Shake() end
	if message then self:Pour(message, 1, 0, 0, nil, 24, "OUTLINE", true) end
end

function Omen:ShowConfig()
	local f = AceConfig.OpenFrames["Omen"]
	AceConfig:Open("Omen", configFrame)
	if not f then
		f = AceConfig.OpenFrames["Omen"]
		f:SetWidth(500)
		f:SetHeight(400)
	end
end

function Omen:ShowDropdown()
	local dd = LibStub("LibDropdown-1.0")
	if dd then
		dd:OpenAce3Menu(self.configOptions)
	end
end
